
/*
 * This file originates from The Android Open Source Project version 4.3_r1
 * and contains some modifications which are (C) COPYRIGHT 2014 ARM Limited
 */

#include "rs_core.rsh"
#include "rs_graphics.rsh"
#include "rs_structs.h"
#include "rs_mali.h"

// ARM - change function name prefix from rs to ml
// ARM - remove unsupported runtime functions
// ARM - use GPU driver's allocation structure.

// Opaque Allocation type operations
extern uint32_t __attribute__((overloadable))
        mlAllocationGetDimX(rs_allocation a) {
    mali_Allocation_t *alloc = (mali_Allocation_t*)a.p;
    return alloc->mHal.state.dimensionX;
}

extern uint32_t __attribute__((overloadable))
        mlAllocationGetDimY(rs_allocation a) {
    mali_Allocation_t *alloc = (mali_Allocation_t*)a.p;
    return alloc->mHal.state.dimensionY;
}

extern uint32_t __attribute__((overloadable))
        mlAllocationGetDimZ(rs_allocation a) {
    mali_Allocation_t *alloc = (mali_Allocation_t*)a.p;
    return alloc->mHal.state.dimensionZ;
}

extern uint32_t __attribute__((overloadable))
        rsMaliAllocationGetDimLOD(rs_allocation a) {
    mali_Allocation_t *alloc = (mali_Allocation_t*)a.p;
    return alloc->mHal.state.hasMipmaps;
}

extern uint32_t __attribute__((overloadable))
        mlAllocationGetDimFaces(rs_allocation a) {
    mali_Allocation_t *alloc = (mali_Allocation_t*)a.p;
    return alloc->mHal.state.hasFaces;
}


extern rs_element __attribute__((overloadable))
        mlAllocationGetElement(rs_allocation a) {
    mali_Allocation_t *alloc = (mali_Allocation_t *)a.p;
    if (alloc == NULL) {
        rs_element nullElem = {0};
        return nullElem;
    }
    mali_Type_t *type = (mali_Type_t *)&(alloc->mHal.state.type);
    rs_element returnElem = {(const int *)&(type->mHal.state.element)};
    return returnElem;
}

#ifdef RS_DEBUG_RUNTIME
#else
#define ELEMENT_AT(T)                                                   \
    extern void __attribute__((overloadable))                           \
    mlSetElementAt_##T(rs_allocation a, T val, uint32_t x) {            \
        mali_Allocation_t *alloc = (mali_Allocation_t*)a.p;                      \
        uint8_t *p = (uint8_t *)alloc->mHal.drvState.mallocPtr; \
        const uint32_t eSize = sizeof(T);                               \
        *((T*)&p[(eSize * x)]) = val;                                   \
    }                                                                   \
    extern void __attribute__((overloadable))                           \
    mlSetElementAt_##T(rs_allocation a, T val, uint32_t x, uint32_t y) { \
        mali_Allocation_t *alloc = (mali_Allocation_t*)a.p;                      \
        uint8_t *p = (uint8_t *)alloc->mHal.drvState.mallocPtr; \
        const uint32_t eSize = sizeof(T);                               \
        const uint32_t stride = alloc->mHal.drvState.stride;     \
        *((T*)&p[(eSize * x) + (y * stride)]) = val;                    \
    }                                                                   \
    extern void __attribute__((overloadable))                           \
    mlSetElementAt_##T(rs_allocation a, T val, uint32_t x, uint32_t y, uint32_t z) { \
        mali_Allocation_t *alloc = (mali_Allocation_t*)a.p;                      \
        uint8_t *p = (uint8_t *)alloc->mHal.drvState.mallocPtr; \
        const uint32_t stride = alloc->mHal.drvState.stride;     \
        const uint32_t dimY = alloc->mHal.state.dimensionY;         \
        uint8_t *dp = &p[(sizeof(T) * x) + (y * stride) + (z * stride * dimY)]; \
        ((T*)dp)[0] = val;                                        \
    }                                                                   \
    extern T __attribute__((overloadable))                              \
    mlGetElementAt_##T(rs_allocation a, uint32_t x) {                   \
        mali_Allocation_t *alloc = (mali_Allocation_t*)a.p;                      \
        const uint8_t *p = (const uint8_t *)alloc->mHal.drvState.mallocPtr; \
        return *((T*)&p[(sizeof(T) * x)]);                              \
    }                                                                   \
    extern T __attribute__((overloadable))                              \
    mlGetElementAt_##T(rs_allocation a, uint32_t x, uint32_t y) {       \
        mali_Allocation_t *alloc = (mali_Allocation_t*)a.p;                      \
        const uint8_t *p = (const uint8_t *)alloc->mHal.drvState.mallocPtr; \
        const uint32_t stride = alloc->mHal.drvState.stride;     \
        return *((T*)&p[(sizeof(T) * x) + (y * stride)]);               \
    }                                                                   \
    extern T __attribute__((overloadable))                              \
    mlGetElementAt_##T(rs_allocation a, uint32_t x, uint32_t y, uint32_t z) { \
        mali_Allocation_t *alloc = (mali_Allocation_t*)a.p;                      \
        const uint8_t *p = (const uint8_t *)alloc->mHal.drvState.mallocPtr; \
        const uint32_t stride = alloc->mHal.drvState.stride;     \
        const uint32_t dimY = alloc->mHal.state.dimensionY;         \
        const uint8_t *dp = &p[(sizeof(T) * x) + (y * stride) + (z * stride * dimY)]; \
        return ((const T*)dp)[0];                                       \
    }


#endif

ELEMENT_AT(char)
ELEMENT_AT(char2)
ELEMENT_AT(char3)
ELEMENT_AT(char4)
ELEMENT_AT(uchar)
ELEMENT_AT(uchar2)
ELEMENT_AT(uchar3)
ELEMENT_AT(uchar4)
ELEMENT_AT(short)
ELEMENT_AT(short2)
ELEMENT_AT(short3)
ELEMENT_AT(short4)
ELEMENT_AT(ushort)
ELEMENT_AT(ushort2)
ELEMENT_AT(ushort3)
ELEMENT_AT(ushort4)
ELEMENT_AT(int)
ELEMENT_AT(int2)
ELEMENT_AT(int3)
ELEMENT_AT(int4)
ELEMENT_AT(uint)
ELEMENT_AT(uint2)
ELEMENT_AT(uint3)
ELEMENT_AT(uint4)
ELEMENT_AT(long)
ELEMENT_AT(long2)
ELEMENT_AT(long3)
ELEMENT_AT(long4)
ELEMENT_AT(ulong)
ELEMENT_AT(ulong2)
ELEMENT_AT(ulong3)
ELEMENT_AT(ulong4)
ELEMENT_AT(float)
ELEMENT_AT(float2)
ELEMENT_AT(float3)
ELEMENT_AT(float4)
ELEMENT_AT(double)
ELEMENT_AT(double2)
ELEMENT_AT(double3)
ELEMENT_AT(double4)

#undef ELEMENT_AT


extern const void * __attribute__((overloadable))
        mlGetElementAt(rs_allocation a, uint32_t x) {
    mali_Allocation_t *alloc = (mali_Allocation_t*)a.p;
    const uint8_t *p = (const uint8_t *)alloc->mHal.drvState.mallocPtr;
    const uint32_t eSize = alloc->mHal.state.elementSizeBytes;
    return (const void *) &p[eSize * x];
}


extern const void * __attribute__((overloadable))
        mlGetElementAt(rs_allocation a, uint32_t x, uint32_t y) {
    mali_Allocation_t *alloc = (mali_Allocation_t*)a.p;
    const uint8_t *p = (const uint8_t *)alloc->mHal.drvState.mallocPtr;
    const uint32_t eSize = alloc->mHal.state.elementSizeBytes;
    const uint32_t stride = alloc->mHal.drvState.stride;
    return (const void *) &p[(eSize * x) + (y * stride)];
}

extern const void * __attribute__((overloadable))
        mlGetElementAt(rs_allocation a, uint32_t x, uint32_t y, uint32_t z) {
    mali_Allocation_t *alloc = (mali_Allocation_t*)a.p;
    const uint8_t *p = (const uint8_t *)alloc->mHal.drvState.mallocPtr;
    const uint32_t eSize = alloc->mHal.state.elementSizeBytes;
    const uint32_t stride = alloc->mHal.drvState.stride;
    const uint32_t dimY = alloc->mHal.state.dimensionY;
    return (const void *) &p[(eSize * x) + (y * stride) + (z * stride * dimY)];
}
