/*
 * This file originates from The Android Open Source Project version 4.3_r1
 * and contains some modifications which are (C) COPYRIGHT 2014 ARM Limited
 */

#include "rs_core.rsh"
#include "rs_cl.rsh"
#include "rs_types.rsh"
#include "rs_math.rsh"


// ARM - change function name prefix from rs to ml
// ARM - add matrix functions implemented in the CPU driver's maths stub functions
// ARM - use memcpy where appropriate

static void memcpy(void * dst, const void * src, uint32_t len) {
  for(uint32_t i=0; i<len; ++i) {
    ((uint8_t *)dst)[i] = ((uint8_t *)src)[i];
  }
}

// _Z20rsMatrixLoadIdentityP12rs_matrix2x2
extern void __attribute__((overloadable))
mlMatrixLoadIdentity(rs_matrix2x2* m) {
    m->m[0] = 1.f;
    m->m[1] = 0.f;
    m->m[2] = 0.f;
    m->m[3] = 1.f;
}

// _Z20rsMatrixLoadIdentityP12rs_matrix3x3
extern void __attribute__((overloadable))
mlMatrixLoadIdentity(rs_matrix3x3* m) {
    m->m[0] = 1.f;
    m->m[1] = 0.f;
    m->m[2] = 0.f;
    m->m[3] = 0.f;
    m->m[4] = 1.f;
    m->m[5] = 0.f;
    m->m[6] = 0.f;
    m->m[7] = 0.f;
    m->m[8] = 1.f;
}

// _Z20rsMatrixLoadIdentityP12rs_matrix4x4
extern void __attribute__((overloadable))
mlMatrixLoadIdentity(rs_matrix4x4* m) {
    m->m[0] = 1.f;
    m->m[1] = 0.f;
    m->m[2] = 0.f;
    m->m[3] = 0.f;
    m->m[4] = 0.f;
    m->m[5] = 1.f;
    m->m[6] = 0.f;
    m->m[7] = 0.f;
    m->m[8] = 0.f;
    m->m[9] = 0.f;
    m->m[10] = 1.f;
    m->m[11] = 0.f;
    m->m[12] = 0.f;
    m->m[13] = 0.f;
    m->m[14] = 0.f;
    m->m[15] = 1.f;
}

// _Z12rsMatrixLoadP12rs_matrix4x4PKf
extern void __attribute__((overloadable))
mlMatrixLoad(rs_matrix4x4* m, float const* v) {
    memcpy(m->m, v, sizeof(m->m));
}

// _Z12rsMatrixLoadP12rs_matrix3x3PKf
extern void __attribute__((overloadable))
mlMatrixLoad(rs_matrix3x3* m, float const* v) {
    memcpy(m->m, v, sizeof(m->m));
}

// _Z12rsMatrixLoadP12rs_matrix2x2PKf
extern void __attribute__((overloadable))
mlMatrixLoad(rs_matrix2x2* m, float const* v) {
    memcpy(m->m, v, sizeof(m->m));
}

// _Z12rsMatrixLoadP12rs_matrix4x4PKS_
extern void __attribute__((overloadable))
mlMatrixLoad(rs_matrix4x4* m, rs_matrix4x4 const* v) {
    memcpy(m->m, v->m, sizeof(m->m));
}

// _Z12rsMatrixLoadP12rs_matrix4x4PK12rs_matrix3x3
extern void __attribute__((overloadable))
mlMatrixLoad(rs_matrix4x4* m, rs_matrix3x3 const* v) {
    m->m[0] = v->m[0];
    m->m[1] = v->m[1];
    m->m[2] = v->m[2];
    m->m[3] = 0.f;
    m->m[4] = v->m[3];
    m->m[5] = v->m[4];
    m->m[6] = v->m[5];
    m->m[7] = 0.f;
    m->m[8] = v->m[6];
    m->m[9] = v->m[7];
    m->m[10] = v->m[8];
    m->m[11] = 0.f;
    m->m[12] = 0.f;
    m->m[13] = 0.f;
    m->m[14] = 0.f;
    m->m[15] = 1.f;
}

// _Z12rsMatrixLoadP12rs_matrix4x4PK12rs_matrix2x2
extern void __attribute__((overloadable))
mlMatrixLoad(rs_matrix4x4* m, rs_matrix2x2 const* v) {
    m->m[0] = v->m[0];
    m->m[1] = v->m[1];
    m->m[2] = 0.f;
    m->m[3] = 0.f;
    m->m[4] = v->m[2];
    m->m[5] = v->m[3];
    m->m[6] = 0.f;
    m->m[7] = 0.f;
    m->m[8] = 0.f;
    m->m[9] = 0.f;
    m->m[10] = 1.f;
    m->m[11] = 0.f;
    m->m[12] = 0.f;
    m->m[13] = 0.f;
    m->m[14] = 0.f;
    m->m[15] = 1.f;
}

//_Z12rsMatrixLoadP12rs_matrix3x3PKS_
extern void __attribute__((overloadable))
mlMatrixLoad(rs_matrix3x3* m, rs_matrix3x3 const* v) {
    memcpy(m->m, v->m, sizeof(m->m));
}

// _Z12rsMatrixLoadP12rs_matrix2x2PKS_
extern void __attribute__((overloadable))
mlMatrixLoad(rs_matrix2x2* m, rs_matrix2x2 const* v) {
    memcpy(m->m, v->m, sizeof(m->m));
}

// _Z18rsMatrixLoadRotateP12rs_matrix4x4ffff
extern void __attribute__((overloadable))
mlMatrixLoadRotate(rs_matrix4x4* m, float rot, float x, float y, float z) {
    float c, s;
    m->m[3] = 0;
    m->m[7] = 0;
    m->m[11]= 0;
    m->m[12]= 0;
    m->m[13]= 0;
    m->m[14]= 0;
    m->m[15]= 1;
    rot *= (float)(M_PI / 180.0f);
    c = cos(rot);
    s = sin(rot);

    const float len = x*x + y*y + z*z;
    if (len != 1) {
        const float recipLen = 1.f / sqrt(len);
        x *= recipLen;
        y *= recipLen;
        z *= recipLen;
    }
    const float nc = 1.0f - c;
    const float xy = x * y;
    const float yz = y * z;
    const float zx = z * x;
    const float xs = x * s;
    const float ys = y * s;
    const float zs = z * s;
    m->m[ 0] = x*x*nc +  c;
    m->m[ 4] =  xy*nc - zs;
    m->m[ 8] =  zx*nc + ys;
    m->m[ 1] =  xy*nc + zs;
    m->m[ 5] = y*y*nc +  c;
    m->m[ 9] =  yz*nc - xs;
    m->m[ 2] =  zx*nc - ys;
    m->m[ 6] =  yz*nc + xs;
    m->m[10] = z*z*nc +  c;
}

// helper functions
extern void __attribute__((overloadable))
mlMatrixSet(rs_matrix4x4* m, uint32_t row, uint32_t col, float v) {
    m->m[row*4 + col] = v;
}
extern float __attribute__((overloadable))
mlMatrixGet(const rs_matrix4x4* m, uint32_t row, uint32_t col) {
    return m->m[row*4 + col];
}
extern float __attribute__((overloadable))
mlMatrixGet(const rs_matrix3x3* m, uint32_t row, uint32_t col) {
    return m->m[row*3 + col];
}

extern void __attribute__((overloadable))
mlMatrixSet(rs_matrix3x3* m, uint32_t row, uint32_t col, float v) {
    m->m[row*3 + col] = v;
}

extern float __attribute__((overloadable))
mlMatrixGet(const rs_matrix2x2* m, uint32_t row, uint32_t col) {
    return m->m[row*2 + col];
}
extern void __attribute__((overloadable))
mlMatrixSet(rs_matrix2x2* m, uint32_t row, uint32_t col, float v) {
    m->m[row*2 + col] = v;
}

// _Z17rsMatrixLoadScaleP12rs_matrix4x4fff
extern void __attribute__((overloadable))
mlMatrixLoadScale(rs_matrix4x4* m, float x, float y, float z) {
    mlMatrixLoadIdentity(m);
    mlMatrixSet(m, 0, 0, x);
    mlMatrixSet(m, 1, 1, y);
    mlMatrixSet(m, 2, 2, z);
}

// _Z21mlMatrixLoadTranslateP12rs_matrix4x4fff
extern void __attribute__((overloadable))
mlMatrixLoadTranslate(rs_matrix4x4* m, float x, float y, float z) {
    mlMatrixLoadIdentity(m);
    m->m[12] = x;
    m->m[13] = y;
    m->m[14] = z;
}

// _Z20rsMatrixLoadMultiplyP12rs_matrix4x4PKS_S2_
extern void __attribute__((overloadable))
mlMatrixLoadMultiply(rs_matrix4x4* m,
        const rs_matrix4x4 *lhs, const rs_matrix4x4 *rhs) {
    for (int i=0 ; i<4 ; i++) {
        float ri0 = 0;
        float ri1 = 0;
        float ri2 = 0;
        float ri3 = 0;
        for (int j=0 ; j<4 ; j++) {
            const float rhs_ij = mlMatrixGet(rhs,i,j);
            ri0 += mlMatrixGet(lhs,j,0) * rhs_ij;
            ri1 += mlMatrixGet(lhs,j,1) * rhs_ij;
            ri2 += mlMatrixGet(lhs,j,2) * rhs_ij;
            ri3 += mlMatrixGet(lhs,j,3) * rhs_ij;
        }
        mlMatrixSet(m, i,0, ri0);
        mlMatrixSet(m, i,1, ri1);
        mlMatrixSet(m, i,2, ri2);
        mlMatrixSet(m, i,3, ri3);
    }
}

// _Z20rsMatrixLoadMultiplyP12rs_matrix3x3PKS_S2_
extern void __attribute__((overloadable))
mlMatrixLoadMultiply(rs_matrix3x3* m,
        const rs_matrix3x3 *lhs, const rs_matrix3x3 *rhs) {
    for (int i=0 ; i<3 ; i++) {
        float ri0 = 0;
        float ri1 = 0;
        float ri2 = 0;
        for (int j=0 ; j<3 ; j++) {
            const float rhs_ij = mlMatrixGet(rhs, i, j);
            ri0 += mlMatrixGet(lhs, j, 0) * rhs_ij;
            ri1 += mlMatrixGet(lhs, j, 1) * rhs_ij;
            ri2 += mlMatrixGet(lhs, j, 2) * rhs_ij;
        }
        mlMatrixSet(m, i, 0, ri0);
        mlMatrixSet(m, i, 1, ri1);
        mlMatrixSet(m, i, 2, ri2);
    }
}


// _Z20rsMatrixLoadMultiplyP12rs_matrix2x2PKS_S2_
extern void __attribute__((overloadable))
mlMatrixLoadMultiply(rs_matrix2x2* m,
        const rs_matrix2x2 *lhs, const rs_matrix2x2 *rhs) {
    for (int i=0 ; i<2 ; i++) {
        float ri0 = 0;
        float ri1 = 0;
        for (int j=0 ; j<2 ; j++) {
            const float rhs_ij = mlMatrixGet(rhs, i, j);
            ri0 += mlMatrixGet(lhs, j, 0) * rhs_ij;
            ri1 += mlMatrixGet(lhs, j, 1) * rhs_ij;
        }
        mlMatrixSet(m, i, 0, ri0);
        mlMatrixSet(m, i, 1, ri1);
    }
}

// _Z16rsMatrixMultiplyP12rs_matrix4x4PKS_
extern void __attribute__((overloadable))
mlMatrixMultiply(rs_matrix4x4* m, rs_matrix4x4 const* v) {
    rs_matrix4x4 tmp;
    mlMatrixLoadMultiply(&tmp, m, v);
    mlMatrixLoad(m, &tmp);
}

// _Z16rsMatrixMultiplyP12rs_matrix3x3PKS_
extern void __attribute__((overloadable))
mlMatrixMultiply(rs_matrix3x3* m, rs_matrix3x3 const* v) {
    rs_matrix3x3 tmp;
    mlMatrixLoadMultiply(&tmp, m, v);
    mlMatrixLoad(m, &tmp);
}

// _Z16rsMatrixMultiplyP12rs_matrix2x2PKS_
extern void __attribute__((overloadable))
mlMatrixMultiply(rs_matrix2x2* m, rs_matrix2x2 const* v) {
    rs_matrix2x2 tmp;
    mlMatrixLoadMultiply(&tmp, m, v);
    mlMatrixLoad(m, &tmp);
}


// _Z14rsMatrixRotateP12rs_matrix4x4ffff
extern void __attribute__((overloadable))
mlMatrixRotate(rs_matrix4x4* m, float rot, float x, float y, float z) {
    rs_matrix4x4 tmp;
    mlMatrixLoadRotate(&tmp, rot, x, y, z);
    mlMatrixMultiply(m, &tmp);
}

// _Z13rsMatrixScaleP12rs_matrix4x4fff
extern void __attribute__((overloadable))
mlMatrixScale(rs_matrix4x4* m, float x, float y, float z) {
    rs_matrix4x4 tmp;
    mlMatrixLoadScale(&tmp, x, y, z);
    mlMatrixMultiply(m, &tmp);
}

// _Z17rsMatrixTranslateP12rs_matrix4x4fff
extern void __attribute__((overloadable))
mlMatrixTranslate(rs_matrix4x4* m, float x, float y, float z) {
    rs_matrix4x4 tmp;
    mlMatrixLoadTranslate(&tmp, x, y, z);
    mlMatrixMultiply(m, &tmp);
}

// _Z17rsMatrixLoadOrthoP12rs_matrix4x4ffffff
extern void __attribute__((overloadable))
mlMatrixLoadOrtho(rs_matrix4x4* m, float left, float right, float bottom, float top, float near, float far) {
    mlMatrixLoadIdentity(m);
    m->m[0] = 2.f / (right - left);
    m->m[5] = 2.f / (top - bottom);
    m->m[10]= -2.f / (far - near);
    m->m[12]= -(right + left) / (right - left);
    m->m[13]= -(top + bottom) / (top - bottom);
    m->m[14]= -(far + near) / (far - near);
}

// _Z19rsMatrixLoadFrustumP12rs_matrix4x4ffffff
extern void __attribute__((overloadable))
mlMatrixLoadFrustum(rs_matrix4x4* m, float left, float right, float bottom, float top, float near, float far) {
    mlMatrixLoadIdentity(m);
    m->m[0] = 2.f * near / (right - left);
    m->m[5] = 2.f * near / (top - bottom);
    m->m[8] = (right + left) / (right - left);
    m->m[9] = (top + bottom) / (top - bottom);
    m->m[10]= -(far + near) / (far - near);
    m->m[11]= -1.f;
    m->m[14]= -2.f * far * near / (far - near);
    m->m[15]= 0.f;
}

// _Z23rsMatrixLoadPerspectiveP12rs_matrix4x4ffff
extern void __attribute__((overloadable))
mlMatrixLoadPerspective(rs_matrix4x4* m, float fovy, float aspect, float near, float far) {
    float top = near * tan((float) (fovy * M_PI / 360.0f));
    float bottom = -top;
    float left = bottom * aspect;
    float right = top * aspect;
    mlMatrixLoadFrustum(m, left, right, bottom, top, near, far);
}

// _Z15rsMatrixInverseP12rs_matrix4x4
extern bool __attribute__((overloadable))
mlMatrixInverse(rs_matrix4x4* m) {
    rs_matrix4x4 result;

    int i, j;
    for (i = 0; i < 4; ++i) {
        for (j = 0; j < 4; ++j) {
            // computeCofactor for int i, int j
            int c0 = (i+1) % 4;
            int c1 = (i+2) % 4;
            int c2 = (i+3) % 4;
            int r0 = (j+1) % 4;
            int r1 = (j+2) % 4;
            int r2 = (j+3) % 4;

            float minor =
                (m->m[c0 + 4*r0] * (m->m[c1 + 4*r1] * m->m[c2 + 4*r2] - m->m[c1 + 4*r2] * m->m[c2 + 4*r1]))
                - (m->m[c0 + 4*r1] * (m->m[c1 + 4*r0] * m->m[c2 + 4*r2] - m->m[c1 + 4*r2] * m->m[c2 + 4*r0]))
                + (m->m[c0 + 4*r2] * (m->m[c1 + 4*r0] * m->m[c2 + 4*r1] - m->m[c1 + 4*r1] * m->m[c2 + 4*r0]));

            float cofactor = (i+j) & 1 ? -minor : minor;

            result.m[4*i + j] = cofactor;
        }
    }

    // Dot product of 0th column of source and 0th row of result
    float det = m->m[0]*result.m[0] + m->m[4]*result.m[1] +
                 m->m[8]*result.m[2] + m->m[12]*result.m[3];

    if (fabs(det) < 1e-6) {
        return false;
    }

    det = 1.0f / det;
    for (i = 0; i < 16; ++i) {
        m->m[i] = result.m[i] * det;
    }

    return true;
}

// _Z24rsMatrixInverseTransposeP12rs_matrix4x4
extern bool __attribute__((overloadable))
mlMatrixInverseTranspose(rs_matrix4x4* m) {
    rs_matrix4x4 result;

    int i, j;
    for (i = 0; i < 4; ++i) {
        for (j = 0; j < 4; ++j) {
            // computeCofactor for int i, int j
            int c0 = (i+1) % 4;
            int c1 = (i+2) % 4;
            int c2 = (i+3) % 4;
            int r0 = (j+1) % 4;
            int r1 = (j+2) % 4;
            int r2 = (j+3) % 4;

            float minor = (m->m[c0 + 4*r0] * (m->m[c1 + 4*r1] * m->m[c2 + 4*r2] - m->m[c1 + 4*r2] * m->m[c2 + 4*r1]))
                         - (m->m[c0 + 4*r1] * (m->m[c1 + 4*r0] * m->m[c2 + 4*r2] - m->m[c1 + 4*r2] * m->m[c2 + 4*r0]))
                         + (m->m[c0 + 4*r2] * (m->m[c1 + 4*r0] * m->m[c2 + 4*r1] - m->m[c1 + 4*r1] * m->m[c2 + 4*r0]));

            float cofactor = (i+j) & 1 ? -minor : minor;

            result.m[4*j + i] = cofactor;
        }
    }

    // Dot product of 0th column of source and 0th column of result
    float det = m->m[0]*result.m[0] + m->m[4]*result.m[4] +
                 m->m[8]*result.m[8] + m->m[12]*result.m[12];

    if (fabs(det) < 1e-6) {
        return false;
    }

    det = 1.0f / det;
    for (i = 0; i < 16; ++i) {
        m->m[i] = result.m[i] * det;
    }

    return true;
}

// _Z17rsMatrixTransposeP12rs_matrix4x4
extern void __attribute__((overloadable))
mlMatrixTranspose(rs_matrix4x4* m) {
    int i, j;
    float temp;
    for (i = 0; i < 3; ++i) {
        for (j = i + 1; j < 4; ++j) {
            temp = m->m[i*4 + j];
            m->m[i*4 + j] = m->m[j*4 + i];
            m->m[j*4 + i] = temp;
        }
    }
}

// _Z17rsMatrixTransposeP12rs_matrix3x3
extern void __attribute__((overloadable))
mlMatrixTranspose(rs_matrix3x3* m) {
    int i, j;
    float temp;
    for (i = 0; i < 2; ++i) {
        for (j = i + 1; j < 3; ++j) {
            temp = mlMatrixGet(m, i, j);
            mlMatrixSet(m, i, j, mlMatrixGet(m, j, i));
            mlMatrixSet(m, j, i, temp);
        }
    }
}

// _Z17rsMatrixTransposeP12rs_matrix2x2
extern void __attribute__((overloadable))
mlMatrixTranspose(rs_matrix2x2* m) {
    float temp = m->m[1];
    m->m[1] = m->m[2];
    m->m[2] = temp;
}

//extern float
//_Z5fractfPU10AS16776960f(float, __attribute__((address_space(1))) float *);

extern float __attribute__((overloadable))
mlFrac (float val) {
  return fmin(val - floor(val), 0x1.fffffep-1f);
}
