/*
 * This file originates from The Android Open Source Project version 4.4.2_r1
 * and contains some modifications which are (C) COPYRIGHT 2014 ARM Limited
 */

#include "rs_core.rsh"
#include "rs_graphics.rsh"
#include "rs_structs.h"
#include "rs_mali.h"
#include "rs_sample.h"

// ARM - change function name prefix from rs to ml
// ARM - restructured to use vector types

static const void * __attribute__((overloadable))
        mlGetElementAt(rs_allocation a, uint2 uv) {
    return (const void *) mlGetElementAt(a, uv.x, uv.y);
}

static uint32_t wrapI(rs_sampler_value wrap, int32_t coord, int32_t size) {
    if (wrap == RS_SAMPLER_WRAP) {
        coord = coord % size;
        if (coord < 0) {
            coord += size;
        }
    }
    if (wrap == RS_SAMPLER_MIRRORED_REPEAT) {
        coord = coord % (size * 2);
        if (coord < 0) {
            coord = (size * 2) + coord;
        }
        if (coord >= size) {
            coord = (size * 2) - coord;
        }
    }
    return (uint32_t)max(0, min(coord, size - 1));
}


// 565 Conversion bits taken from SkBitmap
#define SK_R16_BITS     5
#define SK_G16_BITS     6
#define SK_B16_BITS     5

#define SK_R16_SHIFT    (SK_B16_BITS + SK_G16_BITS)
#define SK_G16_SHIFT    (SK_B16_BITS)
#define SK_B16_SHIFT    0

#define SK_R16_MASK     ((1 << SK_R16_BITS) - 1)
#define SK_G16_MASK     ((1 << SK_G16_BITS) - 1)
#define SK_B16_MASK     ((1 << SK_B16_BITS) - 1)

#define SkGetPackedR16(color)   (((unsigned)(color) >> SK_R16_SHIFT) & SK_R16_MASK)
#define SkGetPackedG16(color)   (((unsigned)(color) >> SK_G16_SHIFT) & SK_G16_MASK)
#define SkGetPackedB16(color)   (((unsigned)(color) >> SK_B16_SHIFT) & SK_B16_MASK)

static inline unsigned SkR16ToR32(unsigned r) {
    return (r << (8 - SK_R16_BITS)) | (r >> (2 * SK_R16_BITS - 8));
}

static inline unsigned SkG16ToG32(unsigned g) {
    return (g << (8 - SK_G16_BITS)) | (g >> (2 * SK_G16_BITS - 8));
}

static inline unsigned SkB16ToB32(unsigned b) {
    return (b << (8 - SK_B16_BITS)) | (b >> (2 * SK_B16_BITS - 8));
}

#define SkPacked16ToR32(c)      SkR16ToR32(SkGetPackedR16(c))
#define SkPacked16ToG32(c)      SkG16ToG32(SkGetPackedG16(c))
#define SkPacked16ToB32(c)      SkB16ToB32(SkGetPackedB16(c))

static float3 getFrom565(uint16_t color) {
    float3 result;
    result.x = (float)SkPacked16ToR32(color);
    result.y = (float)SkPacked16ToG32(color);
    result.z = (float)SkPacked16ToB32(color);
    return result;
}


static float4 __attribute__((overloadable))
        getMaliBilinearSample(rs_allocation a, float2 weights,
                          uint32_t iPixel, uint32_t next,
                          rs_data_kind dk, rs_data_type dt) {
    MALI_BILINEAR_SAMPLE_BODY()
}

static float4 __attribute__((overloadable))
        getMaliBilinearSample(rs_allocation a, float4 weights,
                          uint2 iPixel, uint2 next,
                          rs_data_kind dk, rs_data_type dt) {
    MALI_BILINEAR_SAMPLE_BODY()
}

static float4  __attribute__((overloadable))
        getMaliNearestSample(rs_allocation a, uint32_t iPixel, rs_data_kind dk,
                         rs_data_type dt) {
    MALI_NEAREST_SAMPLE_BODY()
}

static float4  __attribute__((overloadable))
        getMaliNearestSample(rs_allocation a, uint2 iPixel, rs_data_kind dk,
                         rs_data_type dt) {
    MALI_NEAREST_SAMPLE_BODY()
}

static float4 __attribute__((overloadable))
        mali_sample_LOD_LinearPixel(rs_allocation a, const Type_t *type,
                               rs_data_kind dk, rs_data_type dt,
                               rs_sampler s,
                               float uv) {
    mali_Sampler_t *prog = (mali_Sampler_t *)s.p;
    rs_sampler_value wrapS = prog->mHal.state.wrapS;

    mali_Allocation_t * alloc = (mali_Allocation_t *)a.p;
    int32_t sourceW = alloc->mHal.state.dimensionX;
    float pixelUV = uv * (float)(sourceW);
    int32_t iPixel = (int32_t)(pixelUV);
    float frac = pixelUV - (float)iPixel;

    if (frac < 0.5f) {
        iPixel -= 1;
        frac += 0.5f;
    } else {
        frac -= 0.5f;
    }

    float oneMinusFrac = 1.0f - frac;

    float2 weights;
    weights.x = oneMinusFrac;
    weights.y = frac;

    uint32_t next = wrapI(wrapS, iPixel + 1, sourceW);
    uint32_t location = wrapI(wrapS, iPixel, sourceW);

    return getMaliBilinearSample(a, weights, location, next, dk, dt);
}
static float4 __attribute__((overloadable))
        mali_sample_LOD_LinearPixel(rs_allocation a, const Type_t *type,
                               rs_data_kind dk, rs_data_type dt,
                               rs_sampler s,
                               float2 uv) {

    mali_Sampler_t *prog = (mali_Sampler_t *)s.p;
    rs_sampler_value wrapS = prog->mHal.state.wrapS;
    rs_sampler_value wrapT = prog->mHal.state.wrapT;

    mali_Allocation_t * alloc = (mali_Allocation_t *)a.p;
    int32_t sourceW = alloc->mHal.state.dimensionX;
    int32_t sourceH = alloc->mHal.state.dimensionY;

    float2 dimF;
    dimF.x = (float)(sourceW);
    dimF.y = (float)(sourceH);
    float2 pixelUV = uv * dimF;
    int2 iPixel = convert_int2(pixelUV);


    float2 frac = pixelUV - convert_float2(iPixel);

    if (frac.x < 0.5f) {
        iPixel.x -= 1;
        frac.x += 0.5f;
    } else {
        frac.x -= 0.5f;
    }

    if (frac.y < 0.5f) {
        iPixel.y -= 1;
        frac.y += 0.5f;
    } else {
        frac.y -= 0.5f;
    }


    float2 oneMinusFrac = 1.0f - frac;

    float4 weights;
    weights.x = oneMinusFrac.x * oneMinusFrac.y;
    weights.y = frac.x * oneMinusFrac.y;
    weights.z = oneMinusFrac.x * frac.y;
    weights.w = frac.x * frac.y;

    uint2 next;
    next.x = wrapI(wrapS, iPixel.x + 1, sourceW);
    next.y = wrapI(wrapT, iPixel.y + 1, sourceH);
    uint2 location;
    location.x = wrapI(wrapS, iPixel.x, sourceW);
    location.y = wrapI(wrapT, iPixel.y, sourceH);

    return getMaliBilinearSample(a, weights, location, next, dk, dt);
}

static float4 __attribute__((overloadable))
        mali_sample_LOD_NearestPixel(rs_allocation a, const mali_Type_t *type,
                                rs_data_kind dk, rs_data_type dt,
                                rs_sampler s,
                                float uv) {
    mali_Sampler_t *prog = (mali_Sampler_t *)s.p;
    rs_sampler_value wrapS = prog->mHal.state.wrapS;

    mali_Allocation_t * alloc = (mali_Allocation_t *)a.p;
    int32_t sourceW = alloc->mHal.state.dimensionX;
    int32_t iPixel = (int32_t)(uv * (float)(sourceW));
    uint32_t location = wrapI(wrapS, iPixel, sourceW);

    return getMaliNearestSample(a, location, dk, dt);
}

static float4 __attribute__((overloadable))
        mali_sample_LOD_LinearPixel(rs_allocation a, const mali_Type_t *type,
                               rs_data_kind dk, rs_data_type dt,
                               rs_sampler s,
                               float2 uv) {
    mali_Sampler_t *prog = (mali_Sampler_t *)s.p;
    rs_sampler_value wrapS = prog->mHal.state.wrapS;
    rs_sampler_value wrapT = prog->mHal.state.wrapT;

    mali_Allocation_t * alloc = (mali_Allocation_t *)a.p;
    int32_t sourceW = alloc->mHal.state.dimensionX;
    int32_t sourceH = alloc->mHal.state.dimensionY;

    float2 dimF;
    dimF.x = (float)(sourceW);
    dimF.y = (float)(sourceH);
    float2 pixelUV = uv * dimF;
    int2 iPixel = convert_int2(pixelUV);

    float2 frac = pixelUV - convert_float2(iPixel);

    if (frac.x < 0.5f) {
        iPixel.x -= 1;
        frac.x += 0.5f;
    } else {
        frac.x -= 0.5f;
    }
    if (frac.y < 0.5f) {
        iPixel.y -= 1;
        frac.y += 0.5f;
    } else {
        frac.y -= 0.5f;
    }
    float2 oneMinusFrac = 1.0f - frac;

    float4 weights;
    weights.x = oneMinusFrac.x * oneMinusFrac.y;
    weights.y = frac.x * oneMinusFrac.y;
    weights.z = oneMinusFrac.x * frac.y;
    weights.w = frac.x * frac.y;

    uint2 next;
    next.x = wrapI(wrapS, iPixel.x + 1, sourceW);
    next.y = wrapI(wrapT, iPixel.y + 1, sourceH);
    uint2 location;
    location.x = wrapI(wrapS, iPixel.x, sourceW);
    location.y = wrapI(wrapT, iPixel.y, sourceH);

    return getMaliBilinearSample(a, weights, location, next, dk, dt);
}

static float4 __attribute__((overloadable))
        mali_sample_LOD_NearestPixel(rs_allocation a, const mali_Type_t *type,
                                rs_data_kind dk, rs_data_type dt,
                                rs_sampler s,
                                float2 uv) {
    mali_Sampler_t *prog = (mali_Sampler_t *)s.p;
    rs_sampler_value wrapS = prog->mHal.state.wrapS;
    rs_sampler_value wrapT = prog->mHal.state.wrapT;

    mali_Allocation_t * alloc = (mali_Allocation_t *)a.p;
    int32_t sourceW = alloc->mHal.state.dimensionX;
    int32_t sourceH = alloc->mHal.state.dimensionY;

    float2 dimF;
    dimF.x = (float)(sourceW);
    dimF.y = (float)(sourceH);
    int2 iPixel = convert_int2(uv * dimF);

    uint2 location;
    location.x = wrapI(wrapS, iPixel.x, sourceW);
    location.y = wrapI(wrapT, iPixel.y, sourceH);
    return getMaliNearestSample(a, location, dk, dt);
}

extern const float4 __attribute__((overloadable))
        mlSample(rs_allocation a, rs_sampler s, float uv) {
    MALI_SAMPLE_FUNC_BODY()
}


extern const float4 __attribute__((overloadable))
        mlSample(rs_allocation a, rs_sampler s, float2 uv) {
    MALI_SAMPLE_FUNC_BODY()
}
